"""CBOE Major Indices End of Day fetcher."""

from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal, Optional

import pandas as pd
from openbb_cboe.utils.helpers import (
    TICKER_EXCEPTIONS,
    get_cboe_index_directory,
    get_ticker_info,
)
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.major_indices_historical import (
    MajorIndicesHistoricalData,
    MajorIndicesHistoricalQueryParams,
)
from openbb_provider.utils.helpers import make_request
from pydantic import Field


class CboeMajorIndicesHistoricalQueryParams(MajorIndicesHistoricalQueryParams):
    """CBOE Stock end of day query.

    Source: https://www.cboe.com/
    """

    interval: Optional[Literal["1d", "1m"]] = Field(
        description="Use interval, 1m, for intraday prices during the most recent trading period.",
        default="1d",
    )


class CboeMajorIndicesHistoricalData(MajorIndicesHistoricalData):
    """CBOE Stocks End of Day Data."""

    calls_volume: Optional[float] = Field(
        default=None,
        description="Number of calls traded during the most recent trading period. Only valid if interval is 1m.",
    )
    puts_volume: Optional[float] = Field(
        default=None,
        description="Number of puts traded during the most recent trading period. Only valid if interval is 1m.",
    )
    total_options_volume: Optional[float] = Field(
        default=None,
        description="Total number of options traded during the most recent trading period. Only valid if interval is 1m.",
    )


class CboeMajorIndicesHistoricalFetcher(
    Fetcher[
        CboeMajorIndicesHistoricalQueryParams,
        List[CboeMajorIndicesHistoricalData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints."""

    @staticmethod
    def transform_query(
        params: Dict[str, Any]
    ) -> CboeMajorIndicesHistoricalQueryParams:
        """Transform the query. Setting the start and end dates for a 1 year period."""
        return CboeMajorIndicesHistoricalQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeMajorIndicesHistoricalQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint."""
        # Synbol directories are cached for seven days and are used for error handling and URL generation.
        INDEXES = get_cboe_index_directory().index.to_list()
        query.symbol = query.symbol.upper()
        data = pd.DataFrame()
        if "^" in query.symbol:
            query.symbol = query.symbol.replace("^", "")
        query.interval = (
            "1m" if query.symbol == "NDX" and query.interval == "1d" else query.interval
        )

        now = datetime.now()
        query.start_date = (
            query.start_date if query.start_date else now - timedelta(days=50000)
        )
        query.end_date = query.end_date if query.end_date else now

        if query.symbol not in INDEXES and query.symbol not in TICKER_EXCEPTIONS:
            raise RuntimeError(
                f"The symbol, {query.symbol}, was not found in the CBOE index directory. "
                "Use `index_search()` to find supported indices. If the index is European, try `european_index()`."
            )

        def __generate_historical_prices_url(
            symbol,
            data_type: Optional[Literal["intraday", "historical"]] = "historical",
        ) -> str:
            """Generate the final URL for historical prices data."""
            url: str = (
                f"https://cdn.cboe.com/api/global/delayed_quotes/charts/{data_type}"
            )
            url = (
                url + f"/_{symbol}.json"
                if symbol in TICKER_EXCEPTIONS or symbol in INDEXES
                else url + f"/{symbol}.json"
            )
            return url

        url = (
            __generate_historical_prices_url(query.symbol, "intraday")
            if query.interval == "1m"
            else __generate_historical_prices_url(query.symbol)
        )
        r = make_request(url)

        if r.status_code != 200:
            raise RuntimeError(r.status_code)

        if query.interval == "1d":
            data = (
                pd.DataFrame(r.json()["data"])[
                    ["date", "open", "high", "low", "close", "volume"]
                ]
            ).set_index("date")

            # Fill in missing data from current or most recent trading session.

            today = pd.to_datetime(datetime.now().date())
            if today.weekday() > 4:
                day_minus = today.weekday() - 4
                today = pd.to_datetime(today - timedelta(days=day_minus))
            if today != data.index[-1]:
                _today = pd.Series(get_ticker_info(query.symbol))
                today_df = pd.Series(dtype="object")
                today_df["open"] = round(_today["open"], 2)
                today_df["high"] = round(_today["high"], 2)
                today_df["low"] = round(_today["low"], 2)
                today_df["close"] = round(_today["close"], 2)
                if (
                    query.symbol not in INDEXES
                    and query.symbol not in TICKER_EXCEPTIONS
                ):
                    data = data[data["volume"] > 0]
                    today_df["volume"] = _today["volume"]
                today_df["date"] = today.date()
                today_df = pd.DataFrame(today_df).transpose().set_index("date")

                data = pd.concat([data, today_df], axis=0)

            # If ticker is an index there is no volume data and the types must be set.

            if query.symbol in INDEXES or query.symbol in TICKER_EXCEPTIONS:
                data = data[["open", "high", "low", "close", "volume"]]
                data["open"] = round(data.open.astype(float), 2)
                data["high"] = round(data.high.astype(float), 2)
                data["low"] = round(data.low.astype(float), 2)
                data["close"] = round(data.close.astype(float), 2)
                data["volume"] = 0

            data.index = pd.to_datetime(data.index)
            data = data[data["open"] > 0]

            data = data[
                (data.index >= pd.to_datetime(query.start_date))
                & (data.index <= pd.to_datetime(query.end_date))
            ]

        if query.interval == "1m":
            data_list = r.json()["data"]
            date: List[datetime] = []
            open_: List[float] = []
            high: List[float] = []
            low: List[float] = []
            close: List[float] = []
            volume: List[float] = []
            calls_volume: List[float] = []
            puts_volume: List[float] = []
            total_options_volume: List[float] = []

            for i in range(0, len(data_list)):
                date.append(data_list[i]["datetime"])
                open_.append(data_list[i]["price"]["open"])
                high.append(data_list[i]["price"]["high"])
                low.append(data_list[i]["price"]["low"])
                close.append(data_list[i]["price"]["close"])
                volume.append(data_list[i]["volume"]["stock_volume"])
                calls_volume.append(data_list[i]["volume"]["calls_volume"])
                puts_volume.append(data_list[i]["volume"]["puts_volume"])
                total_options_volume.append(
                    data_list[i]["volume"]["total_options_volume"]
                )
            data = pd.DataFrame()
            data["date"] = pd.to_datetime(date)
            data["open"] = open_
            data["high"] = high
            data["low"] = low
            data["close"] = close
            data["volume"] = volume
            data["calls_volume"] = calls_volume
            data["puts_volume"] = puts_volume
            data["total_options_volume"] = total_options_volume
            data = data.set_index("date").sort_index()
            data.index = data.index.astype(str)
            data = data[data["open"] > 0]

        return data.reset_index().to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeMajorIndicesHistoricalQueryParams, data: List[Dict], **kwargs: Any
    ) -> List[CboeMajorIndicesHistoricalData]:
        """Transform the data to the standard format."""
        return [CboeMajorIndicesHistoricalData.model_validate(d) for d in data]



"""CBOE Index Snapshots fetcher."""

from datetime import datetime
from typing import Any, Dict, List, Optional

import pandas as pd
from openbb_cboe.utils.helpers import (
    EUR_INDEX_COLUMNS,
    US_INDEX_COLUMNS,
    Europe,
    get_cboe_directory,
    get_cboe_index_directory,
)
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.index_snapshots import (
    IndexSnapshotsData,
    IndexSnapshotsQueryParams,
)
from openbb_provider.utils.helpers import make_request
from pydantic import Field


class CboeIndexSnapshotsQueryParams(IndexSnapshotsQueryParams):
    """CBOE Index Snapshots query.  Returns quotes for all indices by region.

    Source: https://www.cboe.com/
    """


class CboeIndexSnapshotsData(IndexSnapshotsData):
    """CBOE Company Search Data."""

    isin: Optional[str] = Field(
        default=None,
        description="ISIN code for the index. Valid only for European indices.",
    )
    last_trade_timestamp: Optional[datetime] = Field(
        default=None, description="Last trade timestamp for the index."
    )


class CboeIndexSnapshotsFetcher(
    Fetcher[
        CboeIndexSnapshotsQueryParams,
        List[CboeIndexSnapshotsData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints"""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeIndexSnapshotsQueryParams:
        """Transform the query"""

        return CboeIndexSnapshotsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeIndexSnapshotsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint"""

        data = pd.DataFrame()

        if query.region == "US":
            r = make_request(
                "https://cdn.cboe.com/api/global/delayed_quotes/quotes/all_us_indices.json"
            )

            if r.status_code != 200:
                raise RuntimeError(r.status_code)

            INDEXES = pd.concat(
                [get_cboe_index_directory(), get_cboe_directory()], axis=0
            )
            data = pd.DataFrame.from_records(r.json()["data"])

            data.rename(columns=US_INDEX_COLUMNS, inplace=True)
            data["symbol"] = data["symbol"].str.replace("^", "")
            data["name"] = ""
            data = data.set_index("symbol")
            for i in data.index:
                if i in INDEXES.index:
                    data.at[i, "name"] = INDEXES.at[i, "name"]

            data["name"] = data["name"].astype(str)
            data["currency"] = "USD"
            data = data[
                [
                    "name",
                    "price",
                    "prev_close",
                    "change",
                    "change_percent",
                    "open",
                    "high",
                    "low",
                    "close",
                    "currency",
                    "last_trade_timestamp",
                ]
            ]

        if query.region == "EU":
            r = make_request(
                "https://cdn.cboe.com/api/global/european_indices/index_quotes/all-indices.json"
            )

            if r.status_code != 200:
                raise RuntimeError(r.status_code)

            data = (
                pd.DataFrame.from_records(r.json()["data"])
                .drop(columns=["symbol"])
                .rename(columns={"index": "symbol"})
                .set_index("symbol")
                .round(2)
            )

            INDEXES = pd.DataFrame(Europe.list_indices()).set_index("symbol")

            for i in data.index:
                data.loc[i, ("isin")] = INDEXES.at[i, "isin"]
                data.loc[i, ("name")] = INDEXES.at[i, "name"]
                data.loc[i, ("currency")] = INDEXES.at[i, "currency"]

            data = data[list(EUR_INDEX_COLUMNS.keys())]
            data.columns = list(EUR_INDEX_COLUMNS.values())

        data["change_percent"] = data["change_percent"].fillna(0.0)

        return data.reset_index().to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeIndexSnapshotsQueryParams, data: dict, **kwargs: Any
    ) -> List[CboeIndexSnapshotsData]:
        """Transform the data to the standard format"""
        for item in data:
            item["name"] = item["name"].replace("/", "-")

        return [CboeIndexSnapshotsData.model_validate(d) for d in data]



"""CBOE Options Chains fetcher."""


from datetime import datetime
from typing import Any, Dict, List, Optional

import pandas as pd
from openbb_cboe.utils.helpers import (
    TICKER_EXCEPTIONS,
    get_cboe_directory,
    get_cboe_index_directory,
)
from openbb_provider.abstract.fetcher import Fetcher
from openbb_provider.standard_models.options_chains import (
    OptionsChainsData,
    OptionsChainsQueryParams,
)
from openbb_provider.utils.helpers import make_request
from pydantic import Field, field_validator


class CboeOptionsChainsQueryParams(OptionsChainsQueryParams):
    """Get the complete options chains for a ticker from CBOE.

    Source: https://www.cboe.com/
    """


class CboeOptionsChainsData(OptionsChainsData):
    """CBOE Options Chains Data."""

    bid_size: int = Field(description="Bid size for the option.")
    ask_size: int = Field(description="Ask size for the option.")
    theoretical: float = Field(description="Theoretical value of the option.")
    open: float = Field(description="Opening price of the option.")
    high: float = Field(description="High price of the option.")
    low: float = Field(description="Low price of the option.")
    last_trade_price: Optional[float] = Field(
        description="Last trade price of the option.", default=None
    )
    tick: Optional[str] = Field(
        description="Whether the last tick was up or down in price.", default=None
    )
    prev_close: float = Field(description="Previous closing price of the option.")
    change: float = Field(description="Change in  price of the option.")
    change_percent: float = Field(description="Change, in percent, of the option.")
    implied_volatility: float = Field(description="Implied volatility of the option.")
    delta: float = Field(description="Delta of the option.")
    gamma: float = Field(description="Gamma of the option.")
    vega: float = Field(description="Vega of the option.")
    theta: float = Field(description="Theta of the option.")
    rho: float = Field(description="Rho of the option.")
    last_trade_timestamp: Optional[datetime] = Field(
        description="Last trade timestamp of the option.", default=None
    )
    dte: int = Field(description="Days to expiration for the option.")

    @field_validator("expiration", mode="before", check_fields=False)
    @classmethod
    def date_validate(cls, v):  # pylint: disable=E0213
        """Return the datetime object from the date string"""
        return datetime.strptime(v, "%Y-%m-%d")


class CboeOptionsChainsFetcher(
    Fetcher[
        CboeOptionsChainsQueryParams,
        List[CboeOptionsChainsData],
    ]
):
    """Transform the query, extract and transform the data from the CBOE endpoints"""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> CboeOptionsChainsQueryParams:
        """Transform the query"""
        return CboeOptionsChainsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: CboeOptionsChainsQueryParams,
        credentials: Optional[Dict[str, str]],
        **kwargs: Any,
    ) -> List[Dict]:
        """Return the raw data from the CBOE endpoint"""
        symbol = query.symbol.upper()

        INDEXES = get_cboe_index_directory()
        SYMBOLS = get_cboe_directory()

        if symbol not in SYMBOLS.index:
            raise RuntimeError(f"{symbol} was not found in the CBOE directory.")

        quotes_url = (
            f"https://cdn.cboe.com/api/global/delayed_quotes/options/_{symbol}.json"
            if symbol in TICKER_EXCEPTIONS or symbol in INDEXES.index
            else f"https://cdn.cboe.com/api/global/delayed_quotes/options/{symbol}.json"
        )

        r = make_request(quotes_url)
        if r.status_code != 200:
            raise RuntimeError(f"No options data found for the symbol, {symbol}.")

        r_json = r.json()
        options_df = pd.DataFrame.from_records(r_json["data"]["options"])

        options_df = options_df.rename(
            columns={
                "option": "contract_symbol",
                "iv": "implied_volatility",
                "theo": "theoretical",
                "last_trade_price": "last_trade_price",
                "last_trade_time": "last_trade_timestamp",
                "percent_change": "change_percent",
                "prev_day_close": "prev_close",
            }
        )

        # Parses the option symbols into columns for expiration, strike, and option_type

        option_df_index = options_df["contract_symbol"].str.extractall(
            r"^(?P<Ticker>\D*)(?P<expiration>\d*)(?P<option_type>\D*)(?P<strike>\d*)"
        )
        option_df_index = option_df_index.reset_index().drop(
            columns=["match", "level_0"]
        )
        option_df_index.option_type = option_df_index.option_type.str.replace(
            "C", "call"
        ).str.replace("P", "put")
        option_df_index.strike = [ele.lstrip("0") for ele in option_df_index.strike]
        option_df_index.strike = pd.Series(option_df_index.strike).astype(float)
        option_df_index.strike = option_df_index.strike * (1 / 1000)
        option_df_index.strike = option_df_index.strike.to_list()
        option_df_index.expiration = [
            ele.lstrip("1") for ele in option_df_index.expiration
        ]
        option_df_index.expiration = pd.DatetimeIndex(
            option_df_index.expiration, yearfirst=True
        ).astype(str)
        option_df_index = option_df_index.drop(columns=["Ticker"])

        # Joins the parsed symbol into the dataframe.

        quotes = option_df_index.join(options_df)

        now = datetime.now()
        temp = pd.DatetimeIndex(quotes.expiration)
        temp_ = (temp - now).days + 1
        quotes["dte"] = temp_

        quotes["last_trade_timestamp"] = (
            pd.to_datetime(quotes["last_trade_timestamp"], format="%Y-%m-%dT%H:%M:%S")
            .fillna(value="-")
            .replace("-", None)
        )
        quotes = quotes.set_index(
            keys=["expiration", "strike", "option_type"]
        ).sort_index()
        quotes["open_interest"] = quotes["open_interest"].astype("int64")
        quotes["volume"] = quotes["volume"].astype("int64")
        quotes["bid_size"] = quotes["bid_size"].astype("int64")
        quotes["ask_size"] = quotes["ask_size"].astype("int64")
        quotes["prev_close"] = round(quotes["prev_close"], 2)
        quotes["change_percent"] = round(quotes["change_percent"], 2)

        return quotes.reset_index().to_dict("records")

    @staticmethod
    def transform_data(
        query: CboeOptionsChainsQueryParams,
        data: dict,
        **kwargs: Any,
    ) -> List[CboeOptionsChainsData]:
        """Transform the data to the standard format"""
        return [CboeOptionsChainsData.model_validate(d) for d in data]
